#  This file is part of parallel-ssh.
#  Copyright (C) 2014-2025 Panos Kittenis.
#  Copyright (C) 2014-2025 parallel-ssh Contributors.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation, version 2.1.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import atexit
import logging
from gevent import joinall, get_hub, sleep
from gevent.pool import Pool
from gevent.server import StreamServer
from queue import Queue
from ssh2.error_codes import LIBSSH2_ERROR_EAGAIN
from threading import Thread, Event

from ...constants import DEFAULT_RETRIES

logger = logging.getLogger(__name__)


class LocalForwarder(Thread):

    def __init__(self):
        """Thread runner for a group of local port forwarding proxies.

        Starts servers in their own gevent hub via thread run target.

        Use ``enqueue`` to create new servers
        and get port to connect to via ``out_q`` once a target has been put into the input queue.

        ``SSHClient`` is the client for the SSH host that will be proxying.

        There can be as many LocalForwarder(s) as needed to scale.

        Relationship of clients to forwarders to servers is:

        One client -> One Forwarder -> Many Servers <-> Many proxy hosts -> one target host per proxy host

        A Forwarder starts servers. Servers communicate with clients directly via Forwarder thread.

        Multiple forwarder threads can be used to scale clients to more threads as number of clients increases causing
        contention in forwarder threads handling proxy connections.
        """
        Thread.__init__(self, daemon=True)
        self.in_q = Queue(1)
        self.out_q = Queue(1)
        self._pool = None
        self._servers = {}
        self._hub = None
        self.started = Event()
        self._cleanup_let = None
        self.shutdown_triggered = Event()

    def _start_server(self):
        client, host, port = self.in_q.get()
        logger.debug("Starting server for %s:%s", host, port)
        server = TunnelServer(client, host, port)
        self._servers[client] = server
        server.start()
        self._get_server_listen_port(server)

    def _get_server_listen_port(self, server):
        while not server.started:
            sleep(0.01)
        local_port = server.listen_port
        self.out_q.put(local_port)

    def enqueue(self, client, host, port):
        """Add target host:port to tunnel via client to queue.

        :param client: The client to connect via.
        :type client: :py:mod:`pssh.clients.native.single.SSHClient`
        :param host: Target host to open connection to.
        :type host: str
        :param port: Target port to connect on.
        :type port: int
        """
        self.in_q.put((client, host, port))

    def shutdown(self):
        """Stop all tunnel servers and shutdown LocalForwarder thread.

        This function will join the current thread and wait for it to shutdown if needed.
        """
        for client, server in self._servers.items():
            server.stop()
        self._servers = {}
        if self.started:
            self.shutdown_triggered.set()
            try:
                self.join()
            except RuntimeError:
                # Re-entry protection
                pass

    def _cleanup_servers_let(self):
        while True:
            self._cleanup_servers()
            sleep(60)

    def _cleanup_servers(self):
        for client in list(self._servers.keys()):
            if client.sock is None or client.sock.closed:
                self.cleanup_server(client)

    def run(self):
        """Thread runner ensures a non-main hub has been created for all subsequent
        greenlets and waits for (client, host, port) tuples to be put into self.in_q.

        A server is created once something is in the queue and only then is the port to connect to
        put into self.out_q.
        """
        self._hub = get_hub()
        assert self._hub.main_hub is False
        self._pool = Pool(1)
        self.started.set()
        self._cleanup_let = self._pool.spawn(self._cleanup_servers_let)
        logger.debug("Hub in server runner is main hub: %s", self._hub.main_hub)
        try:
            while True:
                if self.shutdown_triggered.is_set():
                    return
                if self.in_q.empty():
                    sleep(.01)
                    continue
                self._start_server()
        except Exception:
            logger.exception("Tunnel thread caught exception and will exit:")
            self.shutdown()

    def cleanup_server(self, client):
        """
        Stop server for given proxy client and remove client from this forwarder.
        """
        server = self._servers[client]
        server.stop()
        del self._servers[client]


class TunnelServer(StreamServer):
    """Local port forwarding server for tunneling connections from remote SSH server.

    Accepts connections on an available bind_address port once started and tunnels data
    to/from remote SSH host for each connection.

    TunnelServer.listen_port will return listening port for server on given host once TunnelServer.started is True.
    """

    def __init__(self, client, host, port, bind_address='127.0.0.1',
                 num_retries=DEFAULT_RETRIES):
        self.client = client
        self._pool = Pool()
        StreamServer.__init__(self, (bind_address, 0), self._read_rw, spawn=self._pool)
        self.host = host
        self.port = port
        self.session = client.session
        self._retries = num_retries
        self.bind_address = bind_address
        self.exception = None

    @property
    def listen_port(self):
        return self.socket.getsockname()[1] if self.socket is not None else None

    def _read_rw(self, socket, address):
        local_addr, local_port = address
        logger.debug("Client connected, forwarding %s:%s on"
                     " remote host to %s:%s",
                     self.host, self.port, local_addr, local_port)
        try:
            channel = self._open_channel_retries(
                self.host, self.port, local_port)
        except Exception as ex:
            logger.error("Could not establish channel to %s:%s: %s",
                         self.host, self.port, ex)
            self.exception = ex
            return
        # Channel remains alive while this handler function is alive
        source = self._pool.spawn(self._read_forward_sock, socket, channel)
        dest = self._pool.spawn(self._read_channel, socket, channel)
        logger.debug("Waiting for read/write greenlets")
        self._wait_send_receive_lets(source, dest, channel)

    def _wait_send_receive_lets(self, source, dest, channel):
        try:
            joinall((source, dest), raise_error=True)
        finally:
            logger.debug("Read/Write greenlets for tunnel target %s:%s finished, closing forwarding channel",
                         self.host, self.port)
            self.client.close_channel(channel)

    def _read_forward_sock(self, forward_sock, channel):
        while True:
            if channel is None or channel.eof():
                logger.debug("Channel closed, tunnel forward socket reader exiting")
                return
            if not forward_sock or forward_sock.closed:
                logger.debug("Forward socket closed, tunnel forward socket reader exiting")
                return
            try:
                data = forward_sock.recv(1024)
            except Exception as ex:
                logger.error("Forward socket read error: %s", ex)
                raise
            data_len = len(data)
            # logger.debug("Read %s data from forward socket", data_len)
            if data_len == 0:
                sleep(.01)
                continue
            try:
                self.client.eagain_write(channel.write, data)
            except Exception as ex:
                logger.error("Error writing data to channel - %s", ex)
                raise
            logger.debug("Wrote all data to channel")

    def _read_channel(self, forward_sock, channel):
        while True:
            if channel is None or channel.eof():
                logger.debug("Channel closed, tunnel reader exiting")
                return
            try:
                size, data = channel.read()
            except Exception as ex:
                logger.error("Error reading from channel - %s", ex)
                raise
            # logger.debug("Read %s data from channel", size)
            if size == LIBSSH2_ERROR_EAGAIN:
                self.client.poll()
                continue
            elif size == 0:
                sleep(.01)
                continue
            try:
                forward_sock.sendall(data)
            except Exception as ex:
                logger.error(
                    "Error sending data to forward socket - %s", ex)
                raise
            logger.debug("Wrote %s data to forward socket", len(data))

    def _open_channel(self, fw_host, fw_port, local_port):
        channel = self.session.direct_tcpip_ex(
            fw_host, fw_port, self.bind_address,
            local_port)
        while channel == LIBSSH2_ERROR_EAGAIN:
            self.client.poll()
            channel = self.session.direct_tcpip_ex(
                fw_host, fw_port, self.bind_address,
                local_port)
        return channel

    def _open_channel_retries(self, fw_host, fw_port, local_port,
                              wait_time=0.1):
        num_tries = 0
        while num_tries < self._retries:
            try:
                channel = self._open_channel(fw_host, fw_port, local_port)
            except Exception:
                num_tries += 1
                logger.error("Error opening channel to %s:%s, retries %s/%s",
                             fw_host, fw_port, num_tries, self._retries)
                if num_tries >= self._retries:
                    raise
                sleep(wait_time)
                wait_time *= 5
                continue
            return channel


FORWARDER = LocalForwarder()
atexit.register(FORWARDER.shutdown)



#  This file is part of parallel-ssh.
#  Copyright (C) 2014-2025 Panos Kittenis.
#  Copyright (C) 2014-2025 parallel-ssh Contributors.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation, version 2.1.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA


"""Constants definitions for pssh package"""


DEFAULT_RETRIES = 3
"""Default number of retry attempts for SSH client initialisation - authentication, establishing
connections et al."""


RETRY_DELAY = 5
"""Default delay in seconds between retry attempts for SSH client initialisation."""



#  This file is part of parallel-ssh.
#  Copyright (C) 2014-2025 Panos Kittenis.
#  Copyright (C) 2014-2025 parallel-ssh Contributors.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation, version 2.1.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

from io import BytesIO

from gevent.event import Event
from gevent.lock import RLock


class _Eof(Event):
    def __init__(self, unread_data):
        self._unread_data = unread_data
        Event.__init__(self)

    def set(self):
        self._unread_data.set()
        Event.set(self)


class ConcurrentRWBuffer(object):
    """Concurrent reader/writer of bytes for use from multiple greenlets.

    Supports both concurrent reading and writing and combinations there of.

    Iterate on buffer object to read data, yielding event loop if no data exists
    until self.eof has been set.

    Check if end-of-file without blocking with ``ConcurrentRWBuffer.eof.is_set()``.

    Writers should call ``ConcurrentRWBuffer.eof.set()`` when finished writing data via ``write``.

    Readers can use ``read()`` to get any available data or ``None``.
    """
    __slots__ = ('_buffer', '_read_pos', '_write_pos', 'eof', '_lock', '_unread_data')

    def __init__(self):
        self._buffer = BytesIO()
        self._read_pos = 0
        self._write_pos = 0
        self._lock = RLock()
        self._unread_data = Event()
        self.eof = _Eof(self._unread_data)

    def write(self, data):
        """Write data to buffer.

        :param data: Data to write
        :type data: bytes
        """
        with self._lock:
            if not self._buffer.tell() == self._write_pos:
                self._buffer.seek(self._write_pos)
            self._write_pos += self._buffer.write(data)
            if not self._unread_data.is_set() and self._read_pos < self._write_pos:
                self._unread_data.set()

    def read(self):
        """Read available data, or return None.

        :rtype: bytes
        """
        with self._lock:
            if self._write_pos == 0 or self._read_pos == self._write_pos:
                self._unread_data.clear()
                return
            elif not self._buffer.tell() == self._read_pos:
                self._buffer.seek(self._read_pos)
            data = self._buffer.read()
            self._read_pos += len(data)
        return data

    def __iter__(self):
        while not self.eof.is_set() or self._read_pos != self._write_pos:
            data = self.read()
            if data:
                yield data
            else:
                self._unread_data.wait()



#  This file is part of parallel-ssh.
#  Copyright (C) 2014-2025 Panos Kittenis.
#  Copyright (C) 2014-2025 parallel-ssh Contributors.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation, version 2.1.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""Abstract parallel SSH client package"""

import logging

import gevent.pool
from gevent import joinall, spawn, Timeout as GTimeout
from gevent.hub import Hub

from ..common import _validate_pkey_path, _validate_pkey, _validate_api
from ...config import HostConfig
from ...constants import DEFAULT_RETRIES, RETRY_DELAY
from ...exceptions import HostArgumentError, Timeout, ShellError, HostConfigError
from ...output import HostOutput

Hub.NOT_ERROR = (Exception,)
logger = logging.getLogger(__name__)


class BaseParallelSSHClient(object):
    """Parallel client base class."""

    def __init__(self, hosts, user=None, password=None, port=None, pkey=None,
                 allow_agent=True,
                 num_retries=DEFAULT_RETRIES,
                 timeout=120, pool_size=100,
                 host_config=None, retry_delay=RETRY_DELAY,
                 identity_auth=True,
                 ipv6_only=False,
                 proxy_host=None,
                 proxy_port=None,
                 proxy_user=None,
                 proxy_password=None,
                 proxy_pkey=None,
                 keepalive_seconds=None,
                 cert_file=None,
                 gssapi_auth=False,
                 gssapi_server_identity=None,
                 gssapi_client_identity=None,
                 gssapi_delegate_credentials=False,
                 forward_ssh_agent=False,
                 compress=False,
                 keyboard_interactive=False,
                 _auth_thread_pool=True,
                 ):
        self.allow_agent = allow_agent
        self.pool_size = pool_size
        self.pool = gevent.pool.Pool(size=self.pool_size)
        self._hosts = self._validate_hosts(hosts)
        self.user = user
        self.password = password
        self.port = port
        self.pkey = _validate_pkey(pkey)
        self.__pkey_data = self._load_pkey_data(pkey) if pkey is not None else None
        self.num_retries = num_retries
        self.timeout = timeout
        self._host_clients = {}
        self.host_config = host_config
        self.retry_delay = retry_delay
        self.cmds = None
        self.identity_auth = identity_auth
        self.ipv6_only = ipv6_only
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_user = proxy_user
        self.proxy_password = proxy_password
        self.proxy_pkey = proxy_pkey
        self.keepalive_seconds = keepalive_seconds
        self.cert_file = cert_file
        self.forward_ssh_agent = forward_ssh_agent
        self.gssapi_auth = gssapi_auth
        self.gssapi_server_identity = gssapi_server_identity
        self.gssapi_client_identity = gssapi_client_identity
        self.gssapi_delegate_credentials = gssapi_delegate_credentials
        self.compress = compress
        self.keyboard_interactive = keyboard_interactive
        _validate_api(self.keyboard_interactive, self.password)
        self._auth_thread_pool = _auth_thread_pool
        self._check_host_config()

    def _validate_hosts(self, _hosts):
        if _hosts is None:
            raise ValueError
        elif isinstance(_hosts, str) or isinstance(_hosts, bytes):
            raise TypeError(
                "Hosts must be list or other iterable, not string. "
                "For example: ['localhost'] not 'localhost'.")
        elif hasattr(_hosts, '__next__') or hasattr(_hosts, 'next'):
            _hosts = list(_hosts)
        return _hosts

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, _hosts):
        _hosts = self._validate_hosts(_hosts)
        cur_vals = set(enumerate(self._hosts))
        new_vals = set(enumerate(_hosts))
        to_remove = cur_vals.difference(new_vals)
        for i, host in to_remove:
            self._host_clients.pop((i, host), None)
        self._hosts = _hosts

    def _check_host_config(self):
        if self.host_config is None:
            return
        if not isinstance(self.host_config, list):
            raise HostConfigError("Host configuration of type %s is invalid - valid types are List[HostConfig]",
                                  type(self.host_config))
        host_len = len(self.hosts)
        if host_len != len(self.host_config):
            raise ValueError(
                "Host config entries must match number of hosts if provided. "
                "Got %s host config entries from %s hosts" % (
                    len(self.host_config), host_len))

    def _open_shell(self, host_i, host,
                    encoding='utf-8', read_timeout=None):
        try:
            _client = self._get_ssh_client(host_i, host)
            shell = _client.open_shell(
                encoding=encoding, read_timeout=read_timeout)
            return shell
        except (GTimeout, Exception) as ex:
            logger.error("Failed to run on host %s - %s", host, ex)
            raise ex

    def open_shell(self, encoding='utf-8', read_timeout=None):
        """Open interactive shells on all hosts.

        :param encoding: Encoding to use for command string and shell output.
        :type encoding: str
        :param read_timeout: Seconds before reading from output times out.
        :type read_timeout: float

        :returns: Opened shells for each of self.hosts, in order.
        :rtype: list(:py:class:`pssh.clients.native.base.single.InteractiveShell`)
        """
        cmds = [self.pool.spawn(
            self._open_shell, host_i, host, encoding=encoding, read_timeout=read_timeout)
                for host_i, host in enumerate(self.hosts)
                ]
        finished = joinall(cmds, raise_error=True)
        return [cmd.get() for cmd in finished]

    def run_shell_commands(self, shells, commands):
        """Run command(s) on shells.

        :param shells: Shells to run on.
        :type shells: list(:py:class:`pssh.clients.base.single.InteractiveShell`)
        :param commands: Commands to run.
        :type commands: list or str
        """
        if not isinstance(commands, list):
            commands = [commands]
        cmds = [self.pool.spawn(shell.run, cmd)
                for shell in shells
                for cmd in commands]
        try:
            finished = joinall(cmds, raise_error=True, timeout=self.timeout)
        except Exception as ex:
            raise ShellError(ex)
        return finished

    def join_shells(self, shells, timeout=None):
        """Wait for running commands to complete and close shells.

        :param shells: Shells to join on.
        :type shells: list(:py:class:`pssh.clients.base.single.InteractiveShell`)
        :param timeout: Seconds before waiting for shell commands to finish times out.
          Defaults to self.timeout if not provided.
        :type timeout: float

        :raises: :py:class:`pssh.exceptions.Timeout` on timeout requested and
          reached with commands still running.
        """
        _timeout = self.timeout if timeout is None else timeout
        cmds = [self.pool.spawn(shell.close) for shell in shells]
        finished = joinall(cmds, timeout=_timeout)
        if _timeout is None:
            return
        finished_shells = [g.get() for g in finished]
        unfinished_shells = list(set(shells).difference(set(finished_shells)))
        if len(unfinished_shells) > 0:
            raise Timeout(
                "Timeout of %s sec(s) reached with commands still running",
                timeout, finished_shells, unfinished_shells,
            )

    def run_command(self, command, user=None, stop_on_errors=True,
                    host_args=None, use_pty=False, shell=None,
                    encoding='utf-8',
                    *args, **kwargs):
        if host_args:
            try:
                cmds = [self.pool.spawn(
                    self._run_command, host_i, host,
                    command % host_args[host_i],
                    user=user, encoding=encoding,
                    use_pty=use_pty, shell=shell,
                    *args, **kwargs)
                        for host_i, host in enumerate(self.hosts)]
            except IndexError:
                raise HostArgumentError(
                    "Number of host arguments provided does not match "
                    "number of hosts ")
        else:
            cmds = [self.pool.spawn(
                self._run_command, host_i, host, command,
                user=user, encoding=encoding, use_pty=use_pty, shell=shell,
                *args, **kwargs)
                    for host_i, host in enumerate(self.hosts)]
        self.cmds = cmds
        joinall(cmds, timeout=self.timeout)
        return self._get_output_from_cmds(cmds, raise_error=stop_on_errors)

    def _get_output_from_cmds(self, cmds, raise_error=False):
        _cmds = [spawn(self._get_output_from_greenlet, cmd_i, cmd, raise_error=raise_error)
                 for cmd_i, cmd in enumerate(cmds)]
        finished = joinall(_cmds, raise_error=True)
        return [f.get() for f in finished]

    def _get_output_from_greenlet(self, cmd_i, cmd, raise_error=False):
        host = self.hosts[cmd_i]
        alias = self._get_host_config(cmd_i).alias
        try:
            host_out = cmd.get()
            return host_out
        except (GTimeout, Exception) as ex:
            if isinstance(ex, GTimeout):
                ex = Timeout()
            if raise_error:
                raise ex
            return HostOutput(host, None, None, None, exception=ex, alias=alias)

    def get_last_output(self, cmds=None):
        """Get output for last commands executed by ``run_command``.

        :param cmds: Commands to get output for. Defaults to ``client.cmds``
        :type cmds: list(:py:class:`gevent.Greenlet`)

        :rtype: list(:py:class:`pssh.output.HostOutput`)
        """
        cmds = self.cmds if cmds is None else cmds
        if cmds is None:
            return
        return self._get_output_from_cmds(
            cmds, raise_error=False)

    def _get_host_config(self, host_i):
        if self.host_config is None:
            config = HostConfig(
                user=self.user, port=self.port, password=self.password, private_key=self.pkey,
                allow_agent=self.allow_agent, num_retries=self.num_retries, retry_delay=self.retry_delay,
                timeout=self.timeout, identity_auth=self.identity_auth, proxy_host=self.proxy_host,
                proxy_port=self.proxy_port, proxy_user=self.proxy_user, proxy_password=self.proxy_password,
                proxy_pkey=self.proxy_pkey,
                keepalive_seconds=self.keepalive_seconds,
                ipv6_only=self.ipv6_only,
                cert_file=self.cert_file,
                forward_ssh_agent=self.forward_ssh_agent,
                gssapi_auth=self.gssapi_auth,
                gssapi_server_identity=self.gssapi_server_identity,
                gssapi_client_identity=self.gssapi_client_identity,
                gssapi_delegate_credentials=self.gssapi_delegate_credentials,
            )
            return config
        config = self.host_config[host_i]
        return config

    def _run_command(self, host_i, host, command, sudo=False, user=None,
                     shell=None, use_pty=False,
                     encoding='utf-8', read_timeout=None):
        """Make SSHClient if needed, run command on host"""
        try:
            _client = self._get_ssh_client(host_i, host)
            host_out = _client.run_command(
                command, sudo=sudo, user=user, shell=shell,
                use_pty=use_pty, encoding=encoding, read_timeout=read_timeout)
            return host_out
        except (GTimeout, Exception) as ex:
            logger.error("Failed to run on host %s - %s", host, ex)
            raise ex

    def connect_auth(self):
        """Connect to and authenticate with all hosts in parallel.

        This function can be used to perform connection and authentication outside of
        command functions like ``run_command`` or ``copy_file`` so the two operations,
        login and running a remote command, can be separated.

        It is not required to be called prior to any other functions.

        Connections and authentication is performed in parallel by this and all other
        functions.

        :returns: list of greenlets to ``joinall`` with.
        :rtype: list(:py:mod:`gevent.greenlet.Greenlet`)
        """
        cmds = [self.pool.spawn(self._get_ssh_client, i, host) for i, host in enumerate(self.hosts)]
        return cmds

    def _consume_output(self, stdout, stderr):
        for _ in stdout:
            pass
        for _ in stderr:
            pass

    def join(self, output=None, consume_output=False, timeout=None):
        """Wait until all remote commands in output have finished.
        Does *not* block other commands from running in parallel.

        :param output: Output of commands to join on
        :type output: `HostOutput` objects
        :param consume_output: Whether or not join should consume output
          buffers. Output buffers will be empty after ``join`` if set
          to ``True``. Must be set to ``True`` to allow host logger to log
          output on call to ``join`` when host logger has been enabled.
        :type consume_output: bool
        :param timeout: Timeout in seconds if **all** remote commands are not
          yet finished.
          This function's timeout is for all commands in total and will therefor
          be affected by pool size and total number of concurrent commands in
          self.pool.
          Since self.timeout is passed onto each individual SSH session it is
          **not** used for any parallel functions like `run_command` or `join`.
        :type timeout: float

        :raises: :py:class:`pssh.exceptions.Timeout` on timeout requested and
          reached with commands still running.

        :rtype: ``None``"""
        if output is None:
            output = self.get_last_output()
            if output is None:
                logger.info("No last output to join on - run_command has never been run.")
                return
        elif not isinstance(output, list):
            raise ValueError("Unexpected output object type")
        cmds = [self.pool.spawn(self._join, host_out, timeout=timeout,
                                consume_output=consume_output)
                for host_i, host_out in enumerate(output)]
        # Errors raised by self._join should be propagated.
        finished_cmds = joinall(cmds, raise_error=True, timeout=timeout)
        if timeout is None:
            return
        unfinished_cmds = set.difference(set(cmds), set(finished_cmds))
        if unfinished_cmds:
            finished_output = self.get_last_output(cmds=finished_cmds)
            unfinished_output = list(set.difference(set(output), set(finished_output)))
            raise Timeout(
                "Timeout of %s sec(s) reached with commands still running",
                timeout, finished_output, unfinished_output,
            )

    def _join(self, host_out, consume_output=False, timeout=None):
        if host_out is None:
            return
        client = host_out.client
        if client is None:
            return
        client.wait_finished(host_out, timeout=timeout)
        if consume_output:
            self._consume_output(host_out.stdout, host_out.stderr)
        return host_out

    def finished(self, output=None):
        """Check if commands have finished without blocking.

        :param output: (Optional) Output to check if finished. Defaults to
          :py:func:`get_last_output <pssh.clients.base.parallel..ParallelSSHClient.get_last_output>`
        :type output: list(:py:mod:`HostOutput <pssh.output.HostOutput>`)

        :rtype: bool
        """
        if output is None:
            output = self.get_last_output()
            if output is None:
                return True
        for host_out in output:
            chan = host_out.channel
            if host_out.client and not host_out.client.finished(chan):
                return False
        return True

    def copy_file(self, local_file, remote_file, recurse=False, copy_args=None):
        """Copy local file to remote file in parallel

        This function returns a list of greenlets which can be
        `join`-ed on to wait for completion.

        :py:func:`gevent.joinall` function may be used to join on all greenlets
        and will also raise exceptions from them if called with
        ``raise_error=True`` - default is `False`.

        Alternatively call `.get` on each greenlet to raise any exceptions from
        it.

        Exceptions listed here are raised when
        either ``gevent.joinall(<greenlets>, raise_error=True)`` is called
        or ``.get`` is called on each greenlet, not this function itself.

        :param local_file: Local filepath to copy to remote host
        :type local_file: str
        :param remote_file: Remote filepath on remote host to copy file to
        :type remote_file: str
        :param recurse: Whether or not to descend into directories recursively.
        :type recurse: bool
        :param copy_args: (Optional) format local_file and remote_file strings
          with per-host arguments in ``copy_args``. ``copy_args`` length must
          equal length of host list -
          :py:class:`pssh.exceptions.HostArgumentError` is raised otherwise
        :type copy_args: tuple or list

        :rtype: List(:py:class:`gevent.Greenlet`) of greenlets for remote copy
          commands

        :raises: :py:class:`ValueError` when a directory is supplied to
          local_file and recurse is not set
        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
          per-host copy arguments not equal to number of hosts
        :raises: :py:class:`IOError` on I/O errors writing files
        :raises: :py:class:`OSError` on OS errors like permission denied

        .. note ::

          Remote directories in `remote_file` that do not exist will be
          created as long as permissions allow.

        """
        if copy_args:
            try:
                return [self.pool.spawn(self._copy_file, host_i, host,
                                        local_file % copy_args[host_i],
                                        remote_file % copy_args[host_i],
                                        {'recurse': recurse})
                        for host_i, host in enumerate(self.hosts)]
            except IndexError:
                raise HostArgumentError(
                    "Number of per-host copy arguments provided does not match "
                    "number of hosts")
        else:
            return [self.pool.spawn(self._copy_file, host_i, host, local_file,
                                    remote_file, {'recurse': recurse})
                    for host_i, host in enumerate(self.hosts)]

    def _copy_file(self, host_i, host, local_file, remote_file, recurse=False):
        """Make sftp client, copy file"""
        client = self._get_ssh_client(host_i, host)
        return client.copy_file(
            local_file, remote_file, recurse=recurse)

    def copy_remote_file(self, remote_file, local_file, recurse=False,
                         suffix_separator='_', copy_args=None, **kwargs):
        """Copy remote file(s) in parallel as
        <local_file><suffix_separator><host>

        With a ``local_file`` value of ``myfile`` and default separator ``_``
        the resulting filename will be ``myfile_myhost`` for the file from host
        ``myhost``.

        This function, like :py:func:`ParallelSSHClient.copy_file`, returns a
        list of greenlets which can be `join`-ed on to wait for completion.

        :py:func:`gevent.joinall` function may be used to join on all greenlets
        and will also raise exceptions if called with ``raise_error=True`` -
        default is `False`.

        Alternatively call `.get` on each greenlet to raise any exceptions from
        it.

        Exceptions listed here are raised when
        either ``gevent.joinall(<greenlets>, raise_error=True)`` is called
        or ``.get`` is called on each greenlet, not this function itself.

        :param remote_file: remote filepath to copy to local host
        :type remote_file: str
        :param local_file: local filepath on local host to copy file to
        :type local_file: str
        :param recurse: whether or not to recurse
        :type recurse: bool
        :param suffix_separator: (Optional) Separator string between
          filename and host, defaults to ``_``. For example, for a
          ``local_file`` value of ``myfile`` and default separator the
          resulting filename will be ``myfile_myhost`` for the file from
          host ``myhost``. ``suffix_separator`` has no meaning if
          ``copy_args`` is provided
        :type suffix_separator: str
        :param copy_args: (Optional) Format remote_file and local_file strings
          with per-host arguments in ``copy_args``. ``copy_args`` length must
          equal length of host list -
          :py:class:`pssh.exceptions.HostArgumentError` is raised otherwise
        :type copy_args: tuple or list
        :rtype: list(:py:class:`gevent.Greenlet`) of greenlets for remote copy
          commands
        :raises: :py:class:`ValueError` when a directory is supplied to
          local_file and recurse is not set
        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
          per-host copy arguments not equal to number of hosts
        :raises: :py:class:`IOError` on I/O errors writing files
        :raises: :py:class:`OSError` on OS errors like permission denied

        .. note ::
          Local directories in ``local_file`` that do not exist will be
          created as long as permissions allow.

        .. note ::
          File names will be de-duplicated by appending the hostname to the
          filepath separated by ``suffix_separator``.

        """
        if copy_args:
            try:
                return [self.pool.spawn(
                    self._copy_remote_file, host_i, host,
                    remote_file % copy_args[host_i],
                    local_file % copy_args[host_i], recurse=recurse, **kwargs)
                    for host_i, host in enumerate(self.hosts)]
            except IndexError:
                raise HostArgumentError(
                    "Number of per-host copy arguments provided does not match "
                    "number of hosts")
        else:
            return [self.pool.spawn(
                self._copy_remote_file, host_i, host, remote_file,
                suffix_separator.join([local_file, host]), recurse, **kwargs)
                    for host_i, host in enumerate(self.hosts)]

    def _copy_remote_file(self, host_i, host, remote_file, local_file, recurse,
                          **kwargs):
        """Make sftp client, copy file to local"""
        client = self._get_ssh_client(host_i, host)
        return client.copy_remote_file(
            remote_file, local_file, recurse=recurse, **kwargs)

    def _get_ssh_client(self, host_i, host):
        logger.debug("Make client request for host %s, (host_i, host) in clients: %s",
                     host, (host_i, host) in self._host_clients)
        _client = self._host_clients.get((host_i, host))
        if _client is not None:
            return _client
        cfg = self._get_host_config(host_i)
        _pkey = self.pkey if cfg.private_key is None else cfg.private_key
        _pkey_data = self._load_pkey_data(_pkey)
        _client = self._make_ssh_client(host, cfg, _pkey_data)
        self._host_clients[(host_i, host)] = _client
        return _client

    def _load_pkey_data(self, _pkey):
        if not isinstance(_pkey, str):
            return _pkey
        _pkey = _validate_pkey_path(_pkey)
        with open(_pkey, 'rb') as fh:
            _pkey_data = fh.read()
        return _pkey_data

    def _make_ssh_client(self, host, cfg, _pkey_data):
        raise NotImplementedError



#  This file is part of parallel-ssh.
#  Copyright (C) 2014-2025 Panos Kittenis.
#  Copyright (C) 2014-2025 parallel-ssh Contributors.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation, version 2.1.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import logging

from .single import SSHClient
from ..common import _validate_pkey_path, _validate_pkey
from ..base.parallel import BaseParallelSSHClient
from ...constants import DEFAULT_RETRIES, RETRY_DELAY


logger = logging.getLogger(__name__)


class ParallelSSHClient(BaseParallelSSHClient):
    """ssh-python based parallel client."""

    def __init__(self, hosts, user=None, password=None, port=22, pkey=None,
                 cert_file=None,
                 num_retries=DEFAULT_RETRIES, timeout=None, pool_size=100,
                 allow_agent=True, host_config=None, retry_delay=RETRY_DELAY,
                 forward_ssh_agent=False,
                 gssapi_auth=False,
                 gssapi_server_identity=None,
                 gssapi_client_identity=None,
                 gssapi_delegate_credentials=False,
                 identity_auth=True,
                 ipv6_only=False,
                 compress=False,
                 ):
        """
        :param hosts: Hosts to connect to
        :type hosts: list(str)
        :param user: (Optional) User to login as. Defaults to logged in user
        :type user: str
        :param password: (Optional) Password to use for login. Defaults to
          no password
        :type password: str
        :param port: (Optional) Port number to use for SSH connection. Defaults
          to 22.
        :type port: int
        :param pkey: Private key file path to use. Path must be either absolute
          path or relative to user home directory like ``~/<path>``.
          Bytes type input is used as private key data for authentication.
        :type pkey: str or bytes
        :param cert_file: Public key signed certificate file to use for
          authentication. The corresponding private key must also be provided
          via ``pkey`` parameter.
          For example ``pkey='id_rsa', cert_file='id_rsa-cert.pub'`` for RSA
          signed certificate.
          Path must be absolute or relative to user home directory.
        :type cert_file: str
        :param num_retries: (Optional) Number of connection and authentication
          attempts before the client gives up. Defaults to 3.
        :type num_retries: int
        :param retry_delay: Number of seconds to wait between retries. Defaults
          to :py:class:`pssh.constants.RETRY_DELAY`
        :type retry_delay: int or float
        :param timeout: (Optional) Individual SSH client timeout setting in
          seconds passed on to each SSH client spawned by `ParallelSSHClient`.

          This controls timeout setting of socket operations used for SSH
          sessions *on a per session basis* meaning for each individual
          SSH session.

          Defaults to OS default - usually 60 seconds.

          Parallel functions like `run_command` and `join` have a cummulative
          timeout setting that is separate to and
          not affected by `self.timeout`.
        :type timeout: int or float
        :param pool_size: (Optional) Greenlet pool size. Controls
          concurrency, on how many hosts to execute tasks in parallel.
          Defaults to 100. Overhead in event
          loop will determine how high this can be set to, see scaling guide
          lines in project's readme.
        :type pool_size: int
        :param host_config: (Optional) Per-host configuration for cases where
          not all hosts use the same configuration.
        :type host_config: list(:py:class:`pssh.config.HostConfig`)
        :param allow_agent: (Optional) set to False to disable connecting to
          the system's SSH agent. Currently unused - always off.
        :type allow_agent: bool
        :param identity_auth: (Optional) set to False to disable attempting to
          authenticate with default identity files from
          `pssh.clients.base_ssh_client.BaseSSHClient.IDENTITIES`
        :type identity_auth: bool
        :param forward_ssh_agent: (Optional) Turn on SSH agent forwarding -
          equivalent to `ssh -A` from the `ssh` command line utility.
          Defaults to False if not set.
          Currently unused meaning always off.
        :type forward_ssh_agent: bool
        :param gssapi_server_identity: Set GSSAPI server identity.
        :type gssapi_server_identity: str
        :param gssapi_server_identity: Set GSSAPI client identity.
        :type gssapi_server_identity: str
        :param gssapi_delegate_credentials: Enable/disable server credentials
          delegation.
        :type gssapi_delegate_credentials: bool
        :param ipv6_only: Choose IPv6 addresses only if multiple are available
          for the host or raise NoIPv6AddressFoundError otherwise. Note this will
          disable connecting to an IPv4 address if an IP address is provided instead.
        :type ipv6_only: bool
        :param compress: Enable/Disable compression on the client. Defaults to off.
        :type compress: bool

        :raises: :py:class:`pssh.exceptions.PKeyFileError` on errors finding
          provided private key.
        """
        BaseParallelSSHClient.__init__(
            self, hosts, user=user, password=password, port=port, pkey=pkey,
            allow_agent=allow_agent, num_retries=num_retries,
            timeout=timeout, pool_size=pool_size,
            host_config=host_config, retry_delay=retry_delay,
            identity_auth=identity_auth,
            ipv6_only=ipv6_only,
            compress=compress,
        )
        self.pkey = _validate_pkey(pkey)
        self.cert_file = _validate_pkey_path(cert_file)
        self.forward_ssh_agent = forward_ssh_agent
        self.gssapi_auth = gssapi_auth
        self.gssapi_server_identity = gssapi_server_identity
        self.gssapi_client_identity = gssapi_client_identity
        self.gssapi_delegate_credentials = gssapi_delegate_credentials

    def run_command(self, command, sudo=False, user=None, stop_on_errors=True,
                    use_pty=False, host_args=None, shell=None,
                    encoding='utf-8', read_timeout=None,
                    ):
        """Run command on all hosts in parallel, honoring self.pool_size,
        and return output.

        This function will block until all commands have been received
        by remote servers and then return immediately.

        More explicitly, function will return after connection and
        authentication establishment in the case of on new connections and
        after execute
        commands have been accepted by successfully established SSH channels.

        Any connection and/or authentication exceptions will be raised here
        and need catching *unless* ``run_command`` is called with
        ``stop_on_errors=False`` in which case exceptions are added to
        individual host output instead.

        :param command: Command to run
        :type command: str
        :param sudo: (Optional) Run with sudo. Defaults to False
        :type sudo: bool
        :param user: (Optional) User to run command as. Requires sudo access
          for that user from the logged in user account.
        :type user: str
        :param stop_on_errors: (Optional) Raise exception on errors running
          command. Defaults to True. With stop_on_errors set to False,
          exceptions are instead added to output of `run_command`. See example
          usage below.
        :type stop_on_errors: bool
        :param shell: (Optional) Override shell to use to run command with.
          Defaults to login user's defined shell. Use the shell's command
          syntax, eg `shell='bash -c'` or `shell='zsh -c'`.
        :type shell: str
        :param use_pty: (Optional) Enable/Disable use of pseudo terminal
          emulation. Defaults to ``False``
        :type use_pty: bool
        :param host_args: (Optional) Format command string with per-host
          arguments in ``host_args``. ``host_args`` length must equal length of
          host list - :py:class:`pssh.exceptions.HostArgumentError` is
          raised otherwise
        :type host_args: tuple or list
        :param encoding: Encoding to use for command string and output. Must be valid
          `Python codec <https://docs.python.org/library/codecs.html>`_
        :type encoding: str
        :param read_timeout: (Optional) Timeout in seconds for reading from stdout
          or stderr. Defaults to `self.timeout`. Reading from stdout/stderr will
          raise :py:class:`pssh.exceptions.Timeout`
          after ``timeout`` number seconds if remote output is not ready.
        :type read_timeout: float
        :rtype: list(:py:class:`pssh.output.HostOutput`)

        :raises: :py:class:`pssh.exceptions.AuthenticationError` on
          authentication error
        :raises: :py:class:`pssh.exceptions.UnknownHostError` on DNS
          resolution error
        :raises: :py:class:`pssh.exceptions.ConnectionError` on error
          connecting
        :raises: :py:class:`pssh.exceptions.HostArgumentError` on number of
          host arguments not equal to number of hosts
        :raises: :py:class:`TypeError` on not enough host arguments for cmd
          string format
        :raises: :py:class:`KeyError` on no host argument key in arguments
          dict for cmd string format
        :raises: :py:class:`pssh.exceptions.ProxyError` on errors connecting
          to proxy if a proxy host has been set.
        :raises: :py:class:`pssh.exceptions.Timeout` on timeout starting command.
        :raises: Exceptions from :py:mod:`ssh.exceptions` for all other
          specific errors.
        """
        return BaseParallelSSHClient.run_command(
            self, command, stop_on_errors=stop_on_errors, host_args=host_args,
            user=user, shell=shell, sudo=sudo,
            encoding=encoding, use_pty=use_pty,
            read_timeout=read_timeout,
        )

    def _make_ssh_client(self, host, cfg, _pkey_data):
        _client = SSHClient(
            host, user=cfg.user or self.user, password=cfg.password or self.password, port=cfg.port or self.port,
            pkey=_pkey_data, num_retries=cfg.num_retries or self.num_retries,
            alias=cfg.alias,
            timeout=cfg.timeout or self.timeout,
            allow_agent=cfg.allow_agent or self.allow_agent, retry_delay=cfg.retry_delay or self.retry_delay,
            _auth_thread_pool=cfg.auth_thread_pool or self._auth_thread_pool,
            identity_auth=cfg.identity_auth or self.identity_auth,
            ipv6_only=cfg.ipv6_only or self.ipv6_only,
            gssapi_auth=self.gssapi_auth,
            gssapi_server_identity=self.gssapi_server_identity,
            gssapi_client_identity=self.gssapi_client_identity,
            gssapi_delegate_credentials=self.gssapi_delegate_credentials,
            cert_file=cfg.cert_file,
            compress=cfg.compress or self.compress,
        )
        return _client



#  This file is part of parallel-ssh.
#  Copyright (C) 2014-2025 Panos Kittenis.
#  Copyright (C) 2014-2025 parallel-ssh Contributors.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation, version 2.1.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import logging
import os
from getpass import getuser
from socket import gaierror as sock_gaierror, error as sock_error
from warnings import warn

from gevent import sleep, socket, Timeout as GTimeout
from gevent.hub import Hub
from gevent.pool import Pool
from gevent.select import poll, POLLIN, POLLOUT
from gevent.socket import SHUT_RDWR
from ssh2.exceptions import AgentConnectionError, AgentListIdentitiesError, \
    AgentAuthenticationError, AgentGetIdentityError
from ssh2.utils import find_eol

from ..common import _validate_pkey, _validate_api
from ..reader import ConcurrentRWBuffer
from ...constants import DEFAULT_RETRIES, RETRY_DELAY
from ...exceptions import UnknownHostError, AuthenticationError, \
    ConnectionError, Timeout, NoIPv6AddressFoundError
from ...output import HostOutput, HostOutputBuffers, BufferData

Hub.NOT_ERROR = (Exception,)
host_logger = logging.getLogger('pssh.host_logger')
logger = logging.getLogger(__name__)


class PollMixIn(object):
    """MixIn for co-operative socket polling functionality.

    """
    __slots__ = ('sock',)

    def __init__(self, sock=None):
        self.sock = sock

    def poll(self, timeout=None):
        raise NotImplementedError

    def eagain(self, func, *args, **kwargs):
        raise NotImplementedError

    def eagain_write(self, write_func, data):
        raise NotImplementedError

    def _poll_errcodes(self, directions_func, inbound, outbound):
        directions = directions_func()
        if directions == 0:
            return
        events = 0
        if directions & inbound:
            events = POLLIN
        if directions & outbound:
            events |= POLLOUT
        self._poll_socket(events)

    def _poll_socket(self, events):
        if self.sock is None:
            return
        poller = poll()
        poller.register(self.sock, eventmask=events)
        poller.poll(timeout=1)

    def _eagain_errcode(self, func, eagain, *args, **kwargs):
        ret = func(*args, **kwargs)
        while ret == eagain:
            self.poll()
            ret = func(*args, **kwargs)
            sleep()
        return ret

    def _eagain_write_errcode(self, write_func, data, eagain):
        data_len = len(data)
        total_written = 0
        while total_written < data_len:
            rc, bytes_written = write_func(data[total_written:])
            total_written += bytes_written
            if rc == eagain:
                self.poll()
            sleep()


class Stdin(object):
    """Stdin stream for a channel.

    Handles EAGAIN.

    Provides ``write`` and ``flush`` only.
    """
    __slots__ = ('_channel', '_client')

    def __init__(self, channel, client):
        """
        :param channel: The channel the stdin stream is from.
        :type channel: IO object
        :param client: The SSH client the channel is from.
        :type client: ``BaseSSHClient``
        """
        self._channel = channel
        self._client = client

    def write(self, data):
        """Write to stdin.

        :param data: Data to write.
        :type data: str
        """
        return self._client.eagain(self._channel.write, data)

    def flush(self):
        """Flush pending data written to stdin."""
        if not hasattr(self._channel, "flush"):
            return
        return self._client.eagain(self._channel.flush)


class InteractiveShell(object):
    """
    Run commands on an interactive shell.

    Use as context manager to wait for commands to finish on exit.

    Read from .stdout and stderr once context manager has exited.

    ``InteractiveShell.output`` is a :py:class:`pssh.output.HostOutput` object.
    """
    __slots__ = ('_chan', '_client', 'output', '_encoding')
    _EOL = b'\n'

    def __init__(self, channel, client, encoding='utf-8', read_timeout=None):
        """
        :param channel: The channel to open shell on.
        :type channel: ``ssh2.channel.Channel`` or similar.
        :param client: The SSHClient that opened the channel.
        :type client: :py:class:`BaseSSHClient`
        :param encoding: Encoding to use for command string when calling ``run`` and shell output.
        :type encoding: str
        """
        self._chan = channel
        self._client = client
        self._client._shell(self._chan)
        self._encoding = encoding
        self.output = self._client._make_host_output(
            self._chan, encoding=encoding, read_timeout=read_timeout,
            fully_qualified_command=None,
        )

    @property
    def stdout(self):
        """``self.output.stdout``"""
        return self.output.stdout

    @property
    def stderr(self):
        """``self.output.stderr``"""
        return self.output.stderr

    @property
    def stdin(self):
        """``self.output.stdin``"""
        return self.output.stdin

    @property
    def exit_code(self):
        """``self.output.exit_code``"""
        return self.output.exit_code

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """Wait for shell to finish executing and close channel."""
        if self._chan is None:
            return
        self._client.eagain(self._chan.send_eof)
        self._client.wait_finished(self.output)
        return self

    def run(self, cmd):
        """Run command on interactive shell.

        :param cmd: The command string to run.
          Note that ``\\n`` is appended to every string.
        :type cmd: str
        """
        cmd = cmd.encode(self._encoding) + self._EOL
        self._client.eagain_write(self._chan.write, cmd)


class BaseSSHClient(PollMixIn):

    IDENTITIES = (
        os.path.expanduser('~/.ssh/id_rsa'),
        os.path.expanduser('~/.ssh/id_dsa'),
        os.path.expanduser('~/.ssh/identity'),
        os.path.expanduser('~/.ssh/id_ecdsa'),
        os.path.expanduser('~/.ssh/id_ed25519'),
    )

    def __init__(self, host,
                 user=None, password=None, port=None,
                 pkey=None, alias=None,
                 num_retries=DEFAULT_RETRIES,
                 retry_delay=RETRY_DELAY,
                 allow_agent=True, timeout=None,
                 proxy_host=None,
                 proxy_port=None,
                 _auth_thread_pool=True,
                 identity_auth=True,
                 ipv6_only=False,
                 compress=False,
                 keyboard_interactive=False,
                 ):
        super(PollMixIn, self).__init__()
        self._auth_thread_pool = _auth_thread_pool
        self.host = host
        self.alias = alias
        self.user = user if user else getuser()
        self.password = password
        self.keyboard_interactive = keyboard_interactive
        self.port = port if port else 22
        self.num_retries = num_retries
        self.timeout = timeout if timeout else None
        self.retry_delay = retry_delay
        self.allow_agent = allow_agent
        self.session = None
        self._host = proxy_host if proxy_host else host
        self._port = proxy_port if proxy_port else self.port
        self.pkey = _validate_pkey(pkey)
        self.identity_auth = identity_auth
        self._keepalive_greenlet = None
        self.ipv6_only = ipv6_only
        self.compress = compress
        self.keyboard_interactive = keyboard_interactive
        _validate_api(self.keyboard_interactive, self.password)
        self._pool = Pool()
        self._init()

    def _pkey_from_memory(self, pkey_data):
        raise NotImplementedError

    def _init(self):
        self._connect(self._host, self._port)
        self._init_session()
        self._auth_retry()
        self._keepalive()
        logger.debug("Authentication completed successfully - "
                     "setting session to non-blocking mode")
        self.session.set_blocking(0)

    def _auth_retry(self, retries=1):
        try:
            self.auth()
        except Exception as ex:
            if retries < self.num_retries:
                sleep(self.retry_delay)
                return self._auth_retry(retries=retries+1)
            msg = "Authentication error while connecting to %s:%s - %s - retries %s/%s"
            raise AuthenticationError(msg, self.host, self.port, ex, retries, self.num_retries)

    def disconnect(self):
        """Deprecated and a no-op. Disconnections handled by client de-allocation."""
        warn("Deprecated and a no-op - to be removed in future releases.", DeprecationWarning)

    def _disconnect(self):
        raise NotImplementedError

    def __del__(self):
        try:
            self._disconnect()
        except Exception:
            pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._disconnect()

    def open_shell(self, encoding='utf-8', read_timeout=None):
        """Open interactive shell on new channel.

        Can be used as context manager - ``with open_shell() as shell``.

        :param encoding: Encoding to use for command string and shell output.
        :type encoding: str
        :param read_timeout: Timeout in seconds for reading from output.
        :type read_timeout: float
        """
        chan = self.open_session()
        shell = InteractiveShell(chan, self, encoding=encoding, read_timeout=read_timeout)
        return shell

    def _shell(self, channel):
        raise NotImplementedError

    def _disconnect_eagain(self):
        if self.session is not None and self.sock is not None and not self.sock.closed:
            self.eagain(self.session.disconnect)

    def _connect_init_session_retry(self, retries):
        try:
            self._disconnect_eagain()
        except Exception:
            pass
        self.session = None
        if not self.sock.closed:
            try:
                self.sock.shutdown(SHUT_RDWR)
                self.sock.detach()
            except Exception:
                pass
        self.sock = None
        sleep(self.retry_delay)
        self._connect(self._host, self._port, retries=retries)
        return self._init_session(retries=retries)

    def _get_addr_info(self, host, port):
        addr_info = socket.getaddrinfo(host, port, proto=socket.IPPROTO_TCP)
        if self.ipv6_only:
            filtered = [addr for addr in addr_info if addr[0] is socket.AF_INET6]
            if not filtered:
                raise NoIPv6AddressFoundError(
                    "Requested IPv6 only and no IPv6 addresses found for host %s from "
                    "address list %s", host, [addr for _, _, _, _, addr in addr_info])
            addr_info = filtered
        return addr_info

    def _connect(self, host, port, retries=1):
        try:
            addr_info = self._get_addr_info(host, port)
        except sock_gaierror as ex:
            logger.error("Could not resolve host '%s' - retry %s/%s",
                         host, retries, self.num_retries)
            if retries < self.num_retries:
                sleep(self.retry_delay)
                return self._connect(host, port, retries=retries+1)
            unknown_ex = UnknownHostError("Unknown host %s - %s - retry %s/%s",
                                          host, str(ex.args[1]), retries,
                                          self.num_retries)
            raise unknown_ex from ex
        for i, (family, _type, proto, _, sock_addr) in enumerate(addr_info):
            try:
                return self._connect_socket(family, _type, sock_addr, host, port, retries)
            except ConnectionRefusedError as ex:
                if i+1 == len(addr_info):
                    logger.error("No available addresses from %s", [addr[4] for addr in addr_info])
                    ex.args += (host, port)
                    raise
                continue

    def _connect_socket(self, family, _type, sock_addr, host, port, retries):
        self.sock = socket.socket(family, _type)
        if self.timeout:
            self.sock.settimeout(self.timeout)
        logger.debug("Connecting to %s:%s", host, port)
        try:
            self.sock.connect(sock_addr)
        except ConnectionRefusedError:
            raise
        except sock_error as ex:
            logger.error("Error connecting to host '%s:%s' - retry %s/%s",
                         host, port, retries, self.num_retries)
            while retries < self.num_retries:
                sleep(self.retry_delay)
                return self._connect(host, port, retries=retries+1)
            error_type = ex.args[1] if len(ex.args) > 1 else ex.args[0]
            ex = ConnectionError(
                "Error connecting to host '%s:%s' - %s - retry %s/%s",
                host, port, str(error_type), retries,
                self.num_retries,)
            raise ex

    def _identity_auth(self):
        for identity_file in self.IDENTITIES:
            if not os.path.isfile(identity_file):
                continue
            logger.debug(
                "Trying to authenticate with identity file %s",
                identity_file)
            try:
                self._pkey_file_auth(identity_file, password=self.password)
            except Exception as ex:
                logger.debug(
                    "Authentication with identity file %s failed with %s, "
                    "continuing with other identities",
                    identity_file, ex)
                continue
            else:
                logger.info("Authentication succeeded with identity file %s",
                            identity_file)
                return
        raise AuthenticationError("No authentication methods succeeded")

    def _init_session(self, retries=1):
        raise NotImplementedError

    def _keepalive(self):
        raise NotImplementedError

    def auth(self):
        if self.pkey is not None:
            logger.debug(
                "Proceeding with private key authentication")
            return self._pkey_auth(self.pkey)
        if self.allow_agent:
            try:
                self._agent_auth()
            except (AgentAuthenticationError, AgentConnectionError, AgentGetIdentityError,
                    AgentListIdentitiesError) as ex:
                logger.debug("Agent auth failed with %s "
                             "continuing with other authentication methods", repr(ex))
            except Exception as ex:
                logger.error("Agent auth failed with - %s", repr(ex))
            else:
                logger.debug("Authentication with SSH Agent succeeded")
                return
        if self.identity_auth:
            try:
                return self._identity_auth()
            except AuthenticationError:
                if self.password is None:
                    raise
        if self.password is None:
            msg = "No remaining authentication methods"
            logger.error(msg)
            raise AuthenticationError(msg)
        logger.debug("Private key auth failed or not enabled, trying password")
        self._password_auth()

    def _agent_auth(self):
        raise NotImplementedError

    def _password_auth(self):
        raise NotImplementedError

    def _pkey_auth(self, pkey):
        _pkey = pkey
        if isinstance(pkey, str):
            logger.debug("Private key is provided as str, loading from private key file path")
            with open(pkey, 'rb') as fh:
                _pkey = fh.read()
        elif isinstance(pkey, bytes):
            logger.debug("Private key is provided in bytes, using as private key data")
        return self._pkey_from_memory(_pkey)

    def _pkey_file_auth(self, pkey_file, password=None):
        raise NotImplementedError

    def _open_session(self):
        raise NotImplementedError

    def open_session(self):
        raise NotImplementedError

    def _make_host_output(self, channel, encoding, read_timeout, fully_qualified_command=None):
        _stdout_buffer = ConcurrentRWBuffer()
        _stderr_buffer = ConcurrentRWBuffer()
        _stdout_reader, _stderr_reader = self._make_output_readers(
            channel, _stdout_buffer, _stderr_buffer)
        _stdout_reader.start()
        _stderr_reader.start()
        _buffers = HostOutputBuffers(
            stdout=BufferData(rw_buffer=_stdout_buffer, reader=_stdout_reader),
            stderr=BufferData(rw_buffer=_stderr_buffer, reader=_stderr_reader))
        host_out = HostOutput(
            host=self.host, alias=self.alias, channel=channel, stdin=Stdin(channel, self),
            client=self, encoding=encoding, read_timeout=read_timeout,
            buffers=_buffers, fully_qualified_command=fully_qualified_command,
        )
        return host_out

    def _make_output_readers(self, channel, stdout_buffer, stderr_buffer):
        raise NotImplementedError

    def execute(self, cmd, use_pty=False, channel=None):
        """
        Deprecated - use ``run_command`` instead which returns a ``HostOutput`` object.
        """
        warn("Deprecated - use run_command instead.", DeprecationWarning)
        return self._execute(cmd, use_pty=use_pty, channel=channel)

    def _execute(self, cmd, use_pty=False, channel=None):
        raise NotImplementedError

    def read_stderr(self, stderr_buffer, timeout=None):
        """Read standard error buffer.
        Returns a generator of line by line output.

        :param stderr_buffer: Buffer to read from.
        :type stderr_buffer: :py:class:`pssh.clients.reader.ConcurrentRWBuffer`
        :param timeout: Timeout in seconds - defaults to no timeout.
        :type timeout: int or float
        :rtype: generator
        """
        logger.debug("Reading from stderr buffer, timeout=%s", timeout)
        return self._read_output_buffer(stderr_buffer, timeout=timeout)

    def read_output(self, stdout_buffer, timeout=None):
        """Read standard output buffer.
        Returns a generator of line by line output.

        :param stdout_buffer: Buffer to read from.
        :type stdout_buffer: :py:class:`pssh.clients.reader.ConcurrentRWBuffer`
        :param timeout: Timeout in seconds - defaults to no timeout.
        :type timeout: int or float
        :rtype: generator
        """
        logger.debug("Reading from stdout buffer, timeout=%s", timeout)
        return self._read_output_buffer(stdout_buffer, timeout=timeout)

    def _read_output_buffer(self, _buffer, timeout=None):
        timer = GTimeout(seconds=timeout, exception=Timeout)
        remainder = b""
        remainder_len = 0
        timer.start()
        try:
            for data in _buffer:
                pos = 0
                size = len(data)
                while pos < size:
                    linesep, new_line_pos = find_eol(data, pos)
                    if linesep == -1:
                        remainder += data[pos:]
                        remainder_len = len(remainder)
                        break
                    end_of_line = pos+linesep
                    if remainder_len > 0:
                        line = remainder + data[pos:end_of_line]
                        remainder = b""
                        remainder_len = 0
                    else:
                        line = data[pos:end_of_line]
                    yield line
                    pos += linesep + new_line_pos
            if remainder_len > 0:
                # Finished reading without finding ending linesep
                yield remainder
        finally:
            timer.close()

    def _read_output_to_buffer(self, read_func, _buffer):
        raise NotImplementedError

    def wait_finished(self, host_output, timeout=None):
        raise NotImplementedError

    def close_channel(self, channel):
        raise NotImplementedError

    def get_exit_status(self, channel):
        raise NotImplementedError

    def read_output_buffer(self, output_buffer, prefix=None,
                           callback=None,
                           callback_args=None,
                           encoding='utf-8'):
        """Read from output buffers and log to ``host_logger``.

        :param output_buffer: Iterator containing buffer.
        :type output_buffer: iterator
        :param prefix: String to prefix log output to ``host_logger`` with.
        :type prefix: str
        :param callback: Function to call back once buffer is depleted.
        :type callback: function
        :param callback_args: Arguments for call back function.
        :type callback_args: tuple
        :param encoding: Encoding for output.
        :type encoding: str
        """
        prefix = '' if prefix is None else prefix
        for line in output_buffer:
            output = line.decode(encoding)
            host_logger.info("[%s]%s\t%s", self.host, prefix, output)
            yield output
        if callback:
            callback(*callback_args)

    def run_command(self, command, sudo=False, user=None,
                    use_pty=False, shell=None,
                    encoding='utf-8', timeout=None, read_timeout=None):
        """Run remote command.

        :param command: Command to run.
        :type command: str
        :param sudo: Run command via sudo as super-user.
        :type sudo: bool
        :param user: Run command as user via sudo
        :type user: str
        :param use_pty: Whether or not to obtain a PTY on the channel.
        :type use_pty: bool
        :param shell: (Optional) Override shell to use to run command with.
          Defaults to login user's defined shell. Use the shell's command
          syntax, eg `shell='bash -c'` or `shell='zsh -c'`.
        :type shell: str
        :param encoding: Encoding to use for output. Must be valid
          `Python codec <https://docs.python.org/library/codecs.html>`_
        :type encoding: str
        :param read_timeout: (Optional) Timeout in seconds for reading output.
        :type read_timeout: float
        :param timeout: Deprecated - use read_timeout.

        :rtype: :py:class:`pssh.output.HostOutput`
        """
        # Fast path for no command substitution needed
        if not sudo and not user and not shell:
            _command = command
        else:
            _command = ''
            if sudo and not user:
                _command = 'sudo -S '
            elif user:
                _command = 'sudo -u %s -S ' % (user,)
            _shell = shell if shell else '$SHELL -c'
            _command += "%s '%s'" % (_shell, command,)
        _command = _command.encode(encoding)
        with GTimeout(seconds=self.timeout):
            channel = self._execute(_command, use_pty=use_pty)
        _timeout = read_timeout if read_timeout else timeout
        host_out = self._make_host_output(channel, encoding, _timeout, fully_qualified_command=_command)
        return host_out

    def _make_sftp(self):
        raise NotImplementedError

    def _mkdir(self, sftp, directory):
        raise NotImplementedError

    def copy_file(self, local_file, remote_file, recurse=False,
                  sftp=None):
        raise NotImplementedError

    def _sftp_put(self, remote_fh, local_file):
        raise NotImplementedError

    def sftp_put(self, sftp, local_file, remote_file):
        raise NotImplementedError

    def mkdir(self, sftp, directory):
        raise NotImplementedError

    def _copy_dir(self, local_dir, remote_dir, sftp):
        """Call copy_file on every file in the specified directory, copying
        them to the specified remote directory."""
        self.mkdir(sftp, remote_dir)
        file_list = os.listdir(local_dir)
        for file_name in file_list:
            local_path = os.path.join(local_dir, file_name)
            remote_path = '/'.join([remote_dir, file_name])
            self.copy_file(local_path, remote_path, recurse=True,
                           sftp=sftp)

    def copy_remote_file(self, remote_file, local_file, recurse=False,
                         sftp=None, encoding='utf-8'):
        raise NotImplementedError

    def scp_recv(self, remote_file, local_file, recurse=False, sftp=None,
                 encoding='utf-8'):
        raise NotImplementedError

    def _scp_recv(self, remote_file, local_file):
        raise NotImplementedError

    def _scp_send_dir(self, local_dir, remote_dir, sftp):
        file_list = os.listdir(local_dir)
        for file_name in file_list:
            local_path = os.path.join(local_dir, file_name)
            remote_path = '/'.join([remote_dir, file_name])
            self.scp_send(local_path, remote_path, recurse=True,
                          sftp=sftp)

    def _scp_recv_dir(self, file_list, remote_dir, local_dir, sftp,
                      encoding='utf-8'):
        for file_name in file_list:
            file_name = file_name.decode(encoding)
            if file_name in ('.', '..'):
                continue
            remote_path = os.path.join(remote_dir, file_name)
            local_path = os.path.join(local_dir, file_name)
            logger.debug("Attempting recursive copy from %s:%s to %s",
                         self.host, remote_path, local_path)
            self.scp_recv(remote_path, local_path, sftp=sftp,
                          recurse=True)

    def scp_send(self, local_file, remote_file, recurse=False, sftp=None):
        raise NotImplementedError

    def _scp_send(self, local_file, remote_file):
        raise NotImplementedError

    def _sftp_readdir(self, dir_h):
        for size, buf, attrs in dir_h.readdir():
            for line in buf.splitlines():
                yield line

    def _sftp_openfh(self, open_func, remote_file, *args):
        raise NotImplementedError

    def _sftp_get(self, remote_fh, local_file):
        raise NotImplementedError

    def sftp_get(self, sftp, remote_file, local_file):
        raise NotImplementedError

    def _copy_remote_dir(self, file_list, remote_dir, local_dir, sftp,
                         encoding='utf-8'):
        for file_name in file_list:
            file_name = file_name.decode(encoding)
            if file_name in ('.', '..'):
                continue
            remote_path = os.path.join(remote_dir, file_name)
            local_path = os.path.join(local_dir, file_name)
            self.copy_remote_file(remote_path, local_path, sftp=sftp,
                                  recurse=True, encoding=encoding)

    def _make_local_dir(self, dirpath):
        if os.path.exists(dirpath):
            return
        try:
            os.makedirs(dirpath)
        except OSError:
            logger.error("Unable to create local directory structure for "
                         "directory %s", dirpath)
            raise

    def _remote_paths_split(self, file_path):
        _sep = file_path.rfind('/')
        if _sep > 0:
            return file_path[:_sep]
